class Num:
    """
    Defines an integer field for a JSON schema.

    Attributes:
        name (str): The name of the field.
        description (str, optional): A description of the field.
        enums (list, optional): A list of allowed values for the field.
    """

    def __init__(
        self, name: str, description: str | None = None, enums: list | None = None
    ):
        self.name = name
        self.description = description
        self.enums = enums

    def json(self):
        """
        Returns a dictionary representing the JSON schema for this integer field.

        Returns:
            dict: The JSON schema for this integer field.
        """
        schema = {"type": "integer"}
        if self.description:
            schema["description"] = self.description
        if self.enums:
            schema["enum"] = self.enums
        return {self.name: schema}


class Float:
    """
    Defines a float field for a JSON schema.

    Attributes:
        name (str): The name of the field.
        description (str, optional): A description of the field.
        enums (list, optional): A list of allowed values for the field.
    """

    def __init__(
        self, name: str, description: str | None = None, enums: list | None = None
    ):
        self.name = name
        self.description = description
        self.enums = enums

    def json(self):
        """
        Returns a dictionary representing the JSON schema for this float field.

        Returns:
            dict: The JSON schema for this float field.
        """
        schema = {"type": "number"}
        if self.description:
            schema["description"] = self.description
        if self.enums:
            schema["enum"] = self.enums
        return {self.name: schema}


class Str:
    """
    Defines a string field for a JSON schema.

    Attributes:
        name (str): The name of the field.
        description (str, optional): A description of the field.
        enums (list, optional): A list of allowed values for the field.
    """

    def __init__(
        self, name: str, description: str | None = None, enums: list | None = None
    ):
        self.name = name
        self.description = description
        self.enums = enums

    def json(self):
        """
        Returns a dictionary representing the JSON schema for this string field.

        Returns:
            dict: The JSON schema for this string field.
        """
        schema = {"type": "string"}
        if self.description:
            schema["description"] = self.description
        if self.enums:
            schema["enum"] = self.enums
        return {self.name: schema}


class Bool:
    """
    Defines a boolean field for a JSON schema.

    Attributes:
        name (str): The name of the field.
        description (str, optional): A description of the field.
        enums (list, optional): A list of allowed values for the field.
    """

    def __init__(
        self, name: str, description: str | None = None, enums: list | None = None
    ):
        self.name = name
        self.description = description
        self.enums = enums

    def json(self):
        """
        Returns a dictionary representing the JSON schema for this boolean field.

        Returns:
            dict: The JSON schema for this boolean field.
        """
        schema = {"type": "boolean"}
        if self.description:
            schema["description"] = self.description
        if self.enums:
            schema["enum"] = self.enums
        return {self.name: schema}


class Object:
    """
    Defines a dictionary field for a JSON schema.

    Attributes:
        name (str): The name of the field.
        description (str, optional): A description of the field.
        properties (list, optional): A list of objects representing the properties of the dictionary.
        enums (list, optional): A list of allowed values for the field.
    """

    def __init__(
        self,
        name: str,
        description: str | None = None,
        properties: list = [],
        enums: list | None = None,
        required: list[str] | None = None,
    ):
        self.name = name
        self.description = description
        self.properties = properties
        self.enums = enums
        self.required = required

    def json(self):
        """
        Returns a dictionary representing the JSON schema for this dictionary field.

        Returns:
            dict: The JSON schema for this dictionary field.
        """
        schema: dict = {"type": "object"}
        if self.description:
            schema["description"] = self.description
        if self.properties:
            schema["properties"] = {
                prop.name: next(iter(prop.json().values())) for prop in self.properties
            }
        if self.enums:
            schema["enum"] = self.enums
        if self.required:
            schema["required"] = self.required
        schema["additionalProperties"] = False
        return {self.name: schema}


class Array:
    """
    Defines a list field for a JSON schema.

    Attributes:
        name (str): The name of the field.
        description (str, optional): A description of the field.
        item (object, optional): An object representing the type of items in the list.
        enums (list, optional): A list of allowed values for the field.
    """

    def __init__(
        self,
        name: str,
        description: str | None = None,
        item: Str | Num | Float | Bool | Object | None = None,
        enums: list | None = None,
    ):
        self.name = name
        self.description = description
        self.items = item
        self.enums = enums

    def json(self):
        """
        Returns a dictionary representing the JSON schema for this list field.

        Returns:
            dict: The JSON schema for this list field.
        """
        schema = {"type": "array"}
        if self.description:
            schema["description"] = self.description
        if self.items:
            schema["items"] = next(iter(self.items.json().values()))
        if self.enums:
            schema["enum"] = self.enums
        return {self.name: schema}


class AnyOf:
    """
    Defines an anyOf field for a JSON schema.

    Attributes:
        name (str): The name of the field.
        description (str, optional): A description of the field.
        options (list): A list of objects representing the possible types.
    """

    def __init__(self, name: str, options: list):
        self.name = name
        self.options = options

    def json(self):
        """
        Returns a dictionary representing the JSON schema for this anyOf field.

        Returns:
            dict: The JSON schema for this anyOf field.
        """
        schema = {
            "anyOf": [next(iter(option.json().values())) for option in self.options]
        }
        return {self.name: schema}

class Nullable:
    """
    Makes the direct child of this object nullable

    Attributes:
        child (object): The object to make nullable
    """

    def __init__(self, child: Str | Num | Float | Bool | Object | Array):
        self.child = child
    def json(self):
        schema = self.child.json()
        schema[self.child.name]["type"] = [schema[self.child.name]["type"], "null"] 
        return schema

def tool_def_gen(
    name: str,
    description: str | None = None,
    properties: list = [],
    required: list | None = None,
):
    """
    Defines the JSON schema for an OpenAI function call definition.

    Args:
        name (str): The name of the function.
        description (str, optional): A description of the function.
        properties (list, optional): A list of objects representing the properties of the function.
        required (list, optional): A list of required properties for the function.

    Returns:
        dict: The JSON schema for the OpenAI function call.
    """
    schema = {
        "type": "function",
        "function": {
            "name": name,
            "parameters": {
                "type": "object",
                "properties": {k: v for d in properties for k, v in d.json().items()},
            },
        },
    }
    if description:
        schema["function"]["description"] = description
    if required:
        schema["function"]["parameters"]["required"] = required
    return schema


def structured_outputs_gen(
    name: str, properties: list = [], required: list[str] | None = []
):
    """
    Defines the JSON Schema for the OpenAI structured outputs definition.

    Args:
        name (str): The name of the structured output.
        properties (list, optional): A list of property definitions. Defaults to an empty list.
        required (list[str] | None, optional): A list of required properties. Defaults to an empty list.

    Returns:
        dict: The generated schema for the structured output.
    """
    schema = {
        "name": name,
        "strict": True,
        "schema": {
            "type": "object",
            "properties": {k: v for d in properties for k, v in d.json().items()},
            "required": required,
            "additionalProperties": False,
        },
    }
    return schema



import copy
import json
import os
from openai import OpenAI, NotGiven
from ._utils import find_inner_classes, create_messages_dataframe, CustomConverter, parse_json
import warnings
from litellm import completion as litellm_completion
from litellm import stream_chunk_builder as litellm_stream_chunk_builder
from litellm.cost_calculator import cost_per_token as litellm_cost_per_token
import litellm
import abc
import pandas as pd

litellm.drop_params = True


class Tool(abc.ABC):
    """
    This is a base class for tools in the AIIDE module.
    """

    @abc.abstractmethod
    def __init__(self, parent):
        """
        Initializes a new instance of the TOOL class.

        Parameters:
            parent (object): The parent class.

        """
        pass

    @abc.abstractmethod
    def tool_def(self):
        """
        This method is called every time an LLM call is made.
        """
        return {}

    @abc.abstractmethod
    def main(self):
        """
        The main method of the TOOL class.
        """
        return str("")


class Aiide:
    """
    The AIIDE class is the main class for the AIIDE module
    """
    def setup(
        self,
        system_message: str | None = None,
        model: str = "gpt-4o-mini-2024-07-18",
        temperature: float = 1.0,
        api_key: str | None = None,
        history_openai_format: list | None = None,
        **kwargs
    ):
        """
        Setup the AIIDE instance.
        Takes the following arguments:
        - system_message: The system message for the LLM.
        - model: The model to use for the conversation.
        - temperature: The temperature to use for the conversation.
        - api_key: The API key to use for the conversation.
        - history_openai_format: The history of the conversation in OpenAI format. Useful got migrating from OpenAI to AIIDE.
        - kwargs: Additional arguments that are compatible with the LiteLLM API.
        """
        self._api_key = api_key
        self._setup = True
        self._model = model
        self._temperature = temperature
        self.messages: pd.DataFrame = create_messages_dataframe(history_openai_format)
        
        self.usage = {
            "prompt_tokens": 0.0,
            "completion_tokens": 0.0,
            "usd": 0.0,
        }
        if system_message:
            self.messages.loc[len(self.messages)] = { # type: ignore
                "role": "system",
                "content": system_message,
                "arguments": None,
                "response": None,
            }
        self._kwargs = kwargs

    def structured_ouputs(self):
        """
        Structured Outputs is a feature that ensures the model will always generate responses in a specific format. Return JSON Schema definition for the generation.
        """
        return {}

    def chat(
        self,
        user_message: str | list | dict | None = None,
        completion: str | None = None,
        tools: list | None = None,
        stop_words: list | None = None,
        tool_choice: str = "auto",
        json_mode: bool = False,
    ):
        """
        Conversation with AIIDE.

        Args:
            user_message (str, optional): The message from the user to be processed.
            completion (str, optional): The completion of the Agent.
            tools (list, optional): A list of tool instances to be used in the conversation.
            stop_words (list, optional): A list of words that should be considered as stop words for the agent response.
            tool_choice (str, optional): The strategy for choosing which tool to use. Can be "auto", "none", or "required". Defaults to "auto".
            json_mode (bool, optional): If True, the function will return the response in JSON format. Defaults to False.

        Returns:
            yields dictionary with one of the following schema based on response type\n
            if agent response:
                {
                    "type":"text",
                    "content":"",
                    "delta":""
                }
            if tool call:
                {
                    "type":"tool",
                    "name":"",
                    "arguments":""
                }
            if tool response:
                {
                    "type":"tool_response",
                    "name":"",
                    "arguments":""
                    "response":""
                }
        """
        if json_mode:
            response_format = {"type": "json_schema"}
        else:
            response_format = None
        if not hasattr(self, "_setup"):
            raise Exception("Please call self.setup() in __init__")
        # self.setup()
        if completion:
            if user_message:
                self.messages = pd.concat([
                    self.messages,
                    pd.DataFrame({
                        "role": ["user", "assistant"],
                        "content": [user_message, completion],
                        "arguments": [None, None],
                        "response": [None, None],
                    })
                ])
            else:
                self.messages = pd.concat([
                    self.messages,
                    pd.DataFrame({
                        "role": ["assistant"],
                        "content": [completion],
                        "arguments": [None],
                        "response": [None],
                    })
                ])
        elif user_message:
            self.messages = pd.concat([
                self.messages,
                pd.DataFrame({
                    "role": ["user"],
                    "content": [user_message],
                    "arguments": [None],
                    "response": [None],
                })
            ])

        # print(self.messages.aiide.to_openai_dict())
        while True:
            # getting tools
            if tools and len(tools) > 0:
                __tool_definations = []
                __tool_function_mapping = {}
                for each_tool_instance in tools:
                    __tool_definations.append(each_tool_instance.tool_def())
                    __tool_function_mapping[
                        each_tool_instance.tool_def()["function"]["name"]
                    ] = each_tool_instance

            else:
                __tool_definations = None
                tool_choice = None  # type: ignore
            if json_mode:
                # calling structured_output function
                schema = self.structured_ouputs()
                if schema != {}:
                    response_format["json_schema"] = schema  # type: ignore
                    # response_format["strict"] = True
            messages_prev = copy.deepcopy(self.messages.aiide.to_openai_dict())
            response_generator = litellm_completion(
                model=self._model,
                messages=self.messages.aiide.to_openai_dict(),
                tools=__tool_definations,
                tool_choice=tool_choice,  # auto is default, but we'll be explicit
                stream=True,
                temperature=self._temperature,
                stop=stop_words,
                response_format=response_format,
                api_key=self._api_key,
                # adding kwargs
                **self._kwargs,
                # max_tokens=4096,
                # parallel_tool_calls=True,
            )
            response_text = ""
            temp_function_call = []
            chunks = []
            for response_chunk in response_generator:
                chunks.append(response_chunk)
                self.messages.reset_index(drop=True, inplace=True)
                deltas = response_chunk.choices[0].delta  # type: ignore
                finish_reason = response_chunk.choices[0].finish_reason  # type: ignore
                # print("deltas",deltas)
                if deltas.content:
                    response_text += deltas.content
                    if json_mode == True and self.structured_ouputs() != {}:
                        yield_response_text = parse_json(response_text)
                    else:
                        yield_response_text = response_text
                    if self.messages.iloc[-1]["role"] == "assistant":
                        self.messages.loc[self.messages.index[-1], "content"] = response_text
                    else:
                        self.messages = pd.concat([
                            self.messages,
                            pd.DataFrame({
                                "role": ["assistant"],
                                "content": [response_text],
                                "arguments": [None],
                                "response": [None],
                            })
                        ])
                    yield {"type": "text", "content": yield_response_text, "delta": deltas.content}
                
                #! Temporarily disabled yielding of tool calls as they are generated
                # # to yield a tool call, we first check if tool calls have been created, and if so, we check if if the model is actively creating a tool call or if it has finished. hopefully, we can simplify this logic in the future
                # if len(temp_function_call) > 0 and ((deltas.tool_calls != None and deltas.tool_calls[0].function.name) or finish_reason!= None):
                #     # adding the tool call row to self.messages
                #     self.messages.loc[len(self.messages)] = { # type: ignore
                #         "role": "tool",
                #         "content": {
                #             "name": temp_function_call[-1]["name"],
                #             "id": temp_function_call[-1]["tool_call_id"],
                #         },
                #         "arguments": temp_function_call[-1]["arguments"],
                #         "response": None,
                #     } 
                #     yield {
                #         "type": "tool_call",
                #         "name": temp_function_call[-1]["name"],
                #         "arguments": temp_function_call[-1]["arguments"],
                #         "finish": bool(finish_reason),
                #     }
                if deltas.tool_calls != None:
                    # print("tool_calls")
                    if deltas.tool_calls[0].function.name:
                        # new function called
                        temp_function_call.append(
                            {
                                "tool_call_id": deltas.tool_calls[0].id,
                                "name": deltas.tool_calls[0].function.name,
                                "arguments": "",
                            }
                        )
                    if deltas.tool_calls[0].function.arguments != "":
                        # print("adding arguments", deltas.tool_calls[0].function.arguments)
                        temp_function_call[-1]["arguments"] += deltas.tool_calls[0].function.arguments

                if finish_reason:  # type: ignore
                    # print("finish_reason", finish_reason)
                    if finish_reason == "tool_calls":  # type: ignore
                        # calling functions
                        # print("calling funcs", temp_function_call)

                        for tool_index,each_func_call in enumerate(temp_function_call):
                            self.messages.reset_index(drop=True, inplace=True)
                            # adding the tool call row to self.messages
                            self.messages.loc[len(self.messages)] = { # type: ignore
                                "role": "tool",
                                "content": {
                                    "name": each_func_call["name"],
                                    "id": each_func_call["tool_call_id"],
                                },
                                "arguments": each_func_call["arguments"],
                                "response": None,
                            } 
                            yield {
                                "type": "tool_call",
                                "name": each_func_call["name"],
                                "arguments": each_func_call["arguments"],
                                "finish": True if tool_index == len(temp_function_call)-1 else False,
                            }

                            function_to_call = __tool_function_mapping[each_func_call["name"]].main # type: ignore
                            try:
                                function_args = json.loads(each_func_call["arguments"])
                                function_response = function_to_call(**function_args)
                            except Exception as e:
                                # remove prefix string upto first () from error message
                                e = str(e).split(')', 1)[1]
                                function_response = ("Error in function call:\n"+ str(e)+ "\nPlease call the function with the correct format of arguments.")
                            # finding the tool call row in the self.messages dataframe and adding the response
                            # iterating over rows
                            for index, row in self.messages.iterrows():
                                if row["role"] == "tool" and row["content"]["id"] == each_func_call["tool_call_id"]:
                                    # updating the response
                                    self.messages.at[index, "response"] = function_response
                                    break
                            yield {
                                "type": "tool_response",
                                "name": each_func_call["name"],
                                "arguments": each_func_call["arguments"],
                                "response": function_response,
                            }
                        if type(tool_choice) == dict or tool_choice == "required":
                            # If a tool has been forcefully called for more than 100 times, we exit after the final tool execution to avoid usage blowup
                                # warnings.warn("Tools have been called 100 times consecutively. If this is the expected behaviour, please raise an issue on our GitHub Repository!")
                            self.usage["prompt_tokens"] += litellm_stream_chunk_builder(chunks, messages_prev)['usage']["prompt_tokens"]
                            self.usage["completion_tokens"] += litellm_stream_chunk_builder(chunks, messages_prev)['usage']["completion_tokens"]
                            self.usage["usd"] += sum(litellm_cost_per_token(model=self._model, prompt_tokens=self.usage["prompt_tokens"], completion_tokens=self.usage["completion_tokens"]))
                            return
                    elif finish_reason == "length":  # type: ignore
                        warnings.warn("Output token limit reached. Continuing the generation.")
                    else:
                        # print("!!!!!!!GPT STOP")
                        self.usage["prompt_tokens"] += litellm_stream_chunk_builder(chunks, messages_prev)['usage']["prompt_tokens"]
                        self.usage["completion_tokens"] += litellm_stream_chunk_builder(chunks, messages_prev)['usage']["completion_tokens"]
                        self.usage["usd"] += sum(litellm_cost_per_token(model=self._model, prompt_tokens=self.usage["prompt_tokens"], completion_tokens=self.usage["completion_tokens"]))
                        return
            self.usage["prompt_tokens"] += litellm_stream_chunk_builder(chunks, messages_prev)['usage']["prompt_tokens"]
            self.usage["completion_tokens"] += litellm_stream_chunk_builder(chunks, messages_prev)['usage']["completion_tokens"]
            self.usage["usd"] += sum(litellm_cost_per_token(model=self._model, prompt_tokens=self.usage["prompt_tokens"], completion_tokens=self.usage["completion_tokens"]))
        return



import inspect
from PIL import Image
import json
from pandas.api.extensions import register_dataframe_accessor


def find_inner_classes(cls, base_class=None):
    inner_classes = []
    # print("base class", base_class)
    for name, obj in inspect.getmembers(cls):
        if inspect.isclass(obj) and obj.__module__ == cls.__module__:
            # Check if the class is defined inside the given class (not imported)
            if hasattr(obj, "__qualname__") and obj.__qualname__.startswith(
                cls.__qualname__ + "."
            ):
                if base_class is None or issubclass(obj, base_class):
                    inner_classes.append({"class": obj, "type": "agent"})
                else:
                    inner_classes.append({"class": obj, "type": "tool"})

    return inner_classes

def image_to_base64(image):
    import io
    import base64
    from PIL import Image
    buffer = io.BytesIO()
    format = "JPEG"
    image.save(buffer, format=format)
    image_bytes = buffer.getvalue()
    base64_str = base64.b64encode(image_bytes).decode("utf-8")
    return f"data:image/{format.lower()};base64,{base64_str}"

def base64_to_image(base64_image: str):
    from PIL import Image
    import io
    import base64
    base64_image = base64_image.split(",")[1]
    image = Image.open(io.BytesIO(base64.b64decode(base64_image)))
    return image

def create_messages_dataframe(messages):
    import pandas as pd

    df_messages = pd.DataFrame(
        {
            "role": [],
            "content": [],
            "arguments": [],
            "response": [],
        }
    )
    if not messages:
        return df_messages
    # iterate through the messages
    for each_message in messages:
        if "role" in each_message:
            if each_message["role"] == "user" or each_message["role"] == "system":
                if type(each_message["content"]) == str:
                    df_messages = pd.concat(
                        [
                            df_messages,
                            pd.DataFrame(
                                {
                                    "role": [each_message["role"]],
                                    "content": [each_message["content"]],
                                    "arguments": [None],
                                    "response": [None],
                                },
                            ),
                        ]
                    )
                elif type(each_message["content"]) == list:
                    user_content = []
                    for each_content in each_message["content"]:
                        if each_content["type"] == "text":
                            user_content.append(each_content["text"])
                        elif each_content["type"] == "image_url":
                            # converting the image base64 url to a PIL image, "url": f"data:image/jpeg;base64,{base64_image}"
                            # convert the base64 image to a PIL image
                            image = base64_to_image(each_content["url"])

                            user_content.append(image)
                    df_messages = pd.concat(
                        [
                            df_messages,
                            pd.DataFrame(
                                {
                                    "role": [each_message["role"]],
                                    "content": [user_content],
                                    "arguments": [None],
                                    "response": [None],
                                },
                            ),
                        ]
                    )
            elif each_message["role"] == "assistant":
                if each_message["content"]:
                    df_messages = pd.concat(
                        [
                            df_messages,
                            pd.DataFrame(
                                {
                                    "role": [each_message["role"]],
                                    "content": [each_message["content"]],
                                    "arguments": [None],
                                    "response": [None],
                                },
                            ),
                        ]
                    )
                if "tool_calls" in each_message and len(each_message["tool_calls"]) > 0:
                    for each_tool_call in each_message["tool_calls"]:
                        df_messages = pd.concat(
                            [
                                df_messages,
                                pd.DataFrame(
                                    {
                                        "role": ["tool"],
                                        "content": [
                                            {
                                                "id": each_tool_call["id"],
                                                "name": each_tool_call["tool"],
                                            }
                                        ],
                                        "arguments": [each_tool_call["arguments"]],
                                        "response": [each_tool_call["response"]],
                                    },
                                ),
                            ]
                        )
        # reset the index
        df_messages.reset_index(drop=True, inplace=True)
    return df_messages
def content_transformer(content):
    if type(content) == str:
        return content
    elif isinstance(content, Image.Image):
        return [{"type": "image_url", "image_url":{"url": image_to_base64(content)}}]
    
    elif type(content) == list:
        rcontent = []
        for each_content in content:
            if type(each_content) == str:
                rcontent.append({"type": "text", "text": each_content})
            elif isinstance(each_content, Image.Image):
                # convert the PIL image to base64
                rcontent.append(
                    {
                        "type": "image_url",
                        "image_url":{"url": image_to_base64(each_content)},
                    }
                )
        return rcontent
    elif type(content) == dict:
        rcontent = []
        for key, value in content.items():
            if isinstance(value, Image.Image):
                rcontent.append(
                    {
                        "type": "image_url",
                        "image_url": {"url": image_to_base64(value)},
                    }
                )
            elif type(value) == str:
                rcontent.append({"type": "text", "text": value})
            else:
                raise ValueError(f"Invalid type {type(value)}")
        return rcontent

@register_dataframe_accessor("aiide")
class CustomConverter:
    def __init__(self, pandas_obj):
        self.df_messages = pandas_obj  # Reference to the DataFrame
        # reset index
        self.df_messages.reset_index(drop=True, inplace=True)

    def to_openai_dict(self):
        openai_json = {
            "messages": [],
        }
        for index, row in self.df_messages.iterrows():
            if row["role"] == "user" or row["role"] == "system":
                content = content_transformer(row["content"])
                openai_json["messages"].append(
                    {
                        "role": row["role"],
                        "content": content,
                    }
                )
            elif row["role"] == "assistant":
                openai_json["messages"].append(
                    {
                        "role": "assistant",
                        "content": row["content"],
                        "tool_calls": None,
                    }
                )
            elif row["role"] == "tool":
                # finding the index of the last assistant message
                i = 0
                for i in range(len(openai_json["messages"]) - 1, -1, -1):
                    if openai_json["messages"][i]["role"] == "user":
                        i = 0
                        break
                    if openai_json["messages"][i]["role"] == "assistant":
                        break
                if i == 0:
                    # adding an assistant message if there is no assistant message
                    # print("adding an assistant message")
                    openai_json["messages"].append(
                        {
                            "role": "assistant",
                            "content": None,
                            "tool_calls": None,
                        }
                    )
                    i = len(openai_json["messages"]) - 1
                # checking if tool_calls is None
                if openai_json["messages"][i]["tool_calls"] is None:
                    openai_json["messages"][i]["tool_calls"] = []
                openai_json["messages"][i]["tool_calls"].append(
                    {
                        "id": row["content"]["id"],
                        "function": {
                            "name": row["content"]["name"],
                            "arguments": row["arguments"],
                        },
                        "type": "function",
                    }
                )
                openai_json["messages"].append(
                    {
                        "role": "tool",
                        "tool_call_id": row["content"]["id"],
                        "name": row["content"]["name"],
                        "content": content_transformer(row["response"]),
                    }
                )
        return openai_json["messages"]

def parse_json(s, strict=True):
    def on_extra_token(text, data, reminding):
        print('Parsed JSON with extra tokens:', {'text': text, 'data': data, 'reminding': reminding})

    def parse_space(s, e):
        return parse_any(s.strip(), e)

    def parse_array(s, e):
        s = s[1:]  # skip starting '['
        acc = []
        s = s.strip()
        while s:
            if s[0] == ']':
                s = s[1:]  # skip ending ']'
                break
            res, s = parse_any(s, e)
            acc.append(res)
            s = s.strip()
            if s.startswith(','):
                s = s[1:]
                s = s.strip()
        return acc, s

    def parse_object(s, e):
        s = s[1:]  # skip starting '{'
        acc = {}
        s = s.strip()
        while s:
            if s[0] == '}':
                s = s[1:]  # skip ending '}'
                break
            key, s = parse_any(s, e)
            s = s.strip()

            if not s or s[0] == '}':
                acc[key] = None
                break

            if s[0] != ':':
                raise e

            s = s[1:]  # skip ':'
            s = s.strip()

            if not s or s[0] in ',}':
                acc[key] = None
                if s.startswith(','):
                    s = s[1:]
                break

            value, s = parse_any(s, e)
            acc[key] = value
            s = s.strip()
            if s.startswith(','):
                s = s[1:]
                s = s.strip()
        return acc, s

    def parse_string(s, e):
        end = s.find('"', 1)
        while end != -1 and s[end - 1] == '\\':  # Handle escaped quotes
            end = s.find('"', end + 1)
        if end == -1:
            if not strict:
                return s[1:], ""
            else:
                return json.loads(f'"{s[1:]}"'), ""
        str_val = s[:end + 1]
        s = s[end + 1:]
        if not strict:
            return str_val[1:-1], s
        return json.loads(str_val), s

    def parse_number(s, e):
        i = 0
        while i < len(s) and s[i] in '0123456789.-':
            i += 1
        num_str = s[:i]
        s = s[i:]
        if not num_str or num_str == "-" or num_str == ".":
            return num_str, ""
        try:
            if num_str.endswith('.'):
                num = int(num_str[:-1])
            else:
                num = float(num_str) if '.' in num_str or 'e' in num_str or 'E' in num_str else int(num_str)
        except ValueError:
            raise e
        return num, s

    def parse_true(s, e):
        if s.startswith('t') or s.startswith('T'):
            return True, s[4:]
        raise e

    def parse_false(s, e):
        if s.startswith('f') or s.startswith('F'):
            return False, s[5:]
        raise e

    def parse_null(s, e):
        if s.startswith('n'):
            return None, s[4:]
        raise e

    def parse_any(s, e):
        if not s:
            raise e
        parser = parsers.get(s[0])
        if not parser:
            raise e
        return parser(s, e)

    parsers = {
        ' ': parse_space,
        '\r': parse_space,
        '\n': parse_space,
        '\t': parse_space,
        '[': parse_array,
        '{': parse_object,
        '"': parse_string,
        't': parse_true,
        'f': parse_false,
        'n': parse_null
    }
    
    for c in '0123456789.-':
        parsers[c] = parse_number

    if len(s) >= 1:
        try:
            return json.loads(s)
        except json.JSONDecodeError as e:
            data, reminding = parse_any(s, e)
            if on_extra_token and reminding:
                on_extra_token(s, data, reminding)
            return data
    else:
        return json.loads("{}")